import importlib
import importlib.util
import random
import sys
from pathlib import Path

import pytest



def _ensure_rust_extension_importable() -> None:
    """Attempt to expose a locally built Rust extension for test runs."""
    if importlib.util.find_spec("glitchlings._zoo_rust") is not None:
        return

    repo_root = Path(__file__).resolve().parents[1]
    build_root = repo_root / "build"
    if not build_root.exists():
        return

    artifacts = sorted(
        build_root.glob("lib.*/glitchlings/_zoo_rust.*"),
        key=lambda candidate: candidate.stat().st_mtime,
        reverse=True,
    )

    if not artifacts:
        return

    import glitchlings  # Ensure parent package exists before loading extension

    for artifact in artifacts:
        spec = importlib.util.spec_from_file_location("glitchlings._zoo_rust", artifact)
        if spec is None or spec.loader is None:
            continue
        module = importlib.util.module_from_spec(spec)
        sys.modules["glitchlings._zoo_rust"] = module
        spec.loader.exec_module(module)
        package = sys.modules.get("glitchlings")
        if package is not None and hasattr(package, "__path__"):
            package.__path__.append(str(artifact.parent))
        return

_ensure_rust_extension_importable()

reduple_module = importlib.import_module("glitchlings.zoo.reduple")
rushmore_module = importlib.import_module("glitchlings.zoo.rushmore")
scannequin_module = importlib.import_module("glitchlings.zoo.scannequin")
redactyl_module = importlib.import_module("glitchlings.zoo.redactyl")
typogre_module = importlib.import_module("glitchlings.zoo.typogre")
core_module = importlib.import_module("glitchlings.zoo.core")


def _with_descriptor_seeds(
    descriptors: list[dict[str, object]], master_seed: int
) -> list[dict[str, object]]:
    seeded: list[dict[str, object]] = []
    for index, descriptor in enumerate(descriptors):
        seeded.append(
            {
                "name": descriptor["name"],
                "operation": dict(descriptor["operation"]),
                "seed": core_module.Gaggle.derive_seed(
                    master_seed, descriptor["name"], index
                ),
            }
        )
    return seeded


def test_reduple_matches_python_fallback():
    text = "The quick brown fox jumps over the lazy dog."
    expected = reduple_module._python_reduplicate_words(
        text,
        rate=0.5,
        rng=random.Random(123),
    )
    result = reduple_module.reduplicate_words(text, rate=0.5, seed=123)
    assert (
        result
        == expected
        == "The The quick quick brown brown fox fox jumps over over the lazy lazy dog."
    )


def test_reduple_respects_explicit_rng():
    text = "Repeat me"
    expected = reduple_module._python_reduplicate_words(
        text,
        rate=1.0,
        rng=random.Random(99),
    )
    result = reduple_module.reduplicate_words(
        text,
        rate=1.0,
        rng=random.Random(99),
    )
    assert result == expected == "Repeat Repeat me me"


def test_rushmore_matches_python_fallback():
    text = "The quick brown fox jumps over the lazy dog."
    expected = rushmore_module._python_delete_random_words(
        text,
        rate=0.5,
        rng=random.Random(123),
    )
    result = rushmore_module.delete_random_words(
        text, rate=0.5, seed=123
    )
    assert result == expected == "The over the lazy dog."


def test_scannequin_matches_python_fallback():
    text = "The m rn"
    expected = scannequin_module._python_ocr_artifacts(
        text,
        rate=1.0,
        rng=random.Random(1),
    )
    result = scannequin_module.ocr_artifacts(text, rate=1.0, seed=1)
    assert result == expected == "Tlie rn rri"


@pytest.mark.parametrize("seed", [0, 1, 2, 5, 13])
def test_scannequin_overlap_candidates_matches_python(seed: int):
    text = "cl rn li"
    expected = scannequin_module._python_ocr_artifacts(
        text,
        rate=1.0,
        rng=random.Random(seed),
    )
    result = scannequin_module.ocr_artifacts(text, rate=1.0, seed=seed)
    assert result == expected


def test_redactyl_matches_python_fallback():
    text = "The quick brown fox jumps over the lazy dog."
    expected = redactyl_module._python_redact_words(
        text,
        replacement_char=redactyl_module.FULL_BLOCK,
        rate=0.5,
        merge_adjacent=False,
        rng=random.Random(123),
    )
    result = redactyl_module.redact_words(text, rate=0.5, seed=123)
    assert (
        result
        == expected
        == "███ quick brown ███ █████ over the lazy ███."
    )



def test_typogre_matches_python_fallback():
    text = "Adjust the valves before launch"
    layout = getattr(typogre_module.KEYNEIGHBORS, "CURATOR_QWERTY")
    expected = typogre_module._fatfinger_python(
        text,
        rate=0.3,
        layout=layout,
        rng=random.Random(314),
    )
    result = typogre_module.fatfinger(
        text,
        rate=0.3,
        keyboard="CURATOR_QWERTY",
        seed=314,
    )
    assert result == expected

def test_redactyl_merge_adjacent_blocks():
    text = "redact these words"
    expected = redactyl_module._python_redact_words(
        text,
        replacement_char=redactyl_module.FULL_BLOCK,
        rate=1.0,
        merge_adjacent=True,
        rng=random.Random(7),
    )
    result = redactyl_module.redact_words(
        text,
        rate=1.0,
        merge_adjacent=True,
        seed=7,
    )
    assert result == expected == "█████████████████"


def test_redactyl_empty_text_raises_value_error():
    message = "contains no redactable words"
    with pytest.raises(ValueError, match=message):
        redactyl_module.redact_words("", seed=1)


def test_redactyl_whitespace_only_text_raises_value_error():
    message = "contains no redactable words"
    with pytest.raises(ValueError, match=message):
        redactyl_module.redact_words("   \t\n  ", seed=2)


def _run_python_sequence(text: str, descriptors: list[dict[str, object]], master_seed: int) -> str:
    current = text
    for index, descriptor in enumerate(descriptors):
        rng_seed = descriptor.get("seed")
        if rng_seed is None:
            rng_seed = core_module.Gaggle.derive_seed(
                master_seed, descriptor["name"], index
            )
        rng = random.Random(rng_seed)
        operation = descriptor["operation"]
        op_type = operation["type"]
        if op_type == "reduplicate":
            current = reduple_module._python_reduplicate_words(
                current,
                rate=operation["reduplication_rate"],
                rng=rng,
            )
        elif op_type == "delete":
            current = rushmore_module._python_delete_random_words(
                current,
                rate=operation["max_deletion_rate"],
                rng=rng,
            )
        elif op_type == "redact":
            current = redactyl_module._python_redact_words(
                current,
                replacement_char=operation["replacement_char"],
                rate=operation["redaction_rate"],
                merge_adjacent=operation["merge_adjacent"],
                rng=rng,
            )
        elif op_type == "ocr":
            current = scannequin_module._python_ocr_artifacts(
                current,
                rate=operation["error_rate"],
                rng=rng,
            )
        else:  # pragma: no cover - defensive guard
            raise AssertionError(f"Unsupported operation type: {op_type!r}")
    return current


def test_compose_glitchlings_matches_python_pipeline():
    zoo_rust = pytest.importorskip("glitchlings._zoo_rust")
    raw_descriptors = [
        {"name": "Reduple", "operation": {"type": "reduplicate", "reduplication_rate": 0.4}},
        {"name": "Rushmore", "operation": {"type": "delete", "max_deletion_rate": 0.5}},
        {
            "name": "Redactyl",
            "operation": {
                "type": "redact",
                "replacement_char": redactyl_module.FULL_BLOCK,
                "redaction_rate": 0.6,
                "merge_adjacent": True,
            },
        },
        {"name": "Scannequin", "operation": {"type": "ocr", "error_rate": 0.25}},
    ]
    text = "Guard the vault at midnight"
    master_seed = 404
    descriptors = _with_descriptor_seeds(raw_descriptors, master_seed)
    expected = _run_python_sequence(text, descriptors, master_seed)
    result = zoo_rust.compose_glitchlings(text, descriptors, master_seed)
    assert result == expected


def test_compose_glitchlings_is_deterministic():
    zoo_rust = pytest.importorskip("glitchlings._zoo_rust")
    raw_descriptors = [
        {"name": "Reduple", "operation": {"type": "reduplicate", "reduplication_rate": 0.4}},
        {"name": "Rushmore", "operation": {"type": "delete", "max_deletion_rate": 0.3}},
        {
            "name": "Redactyl",
            "operation": {
                "type": "redact",
                "replacement_char": redactyl_module.FULL_BLOCK,
                "redaction_rate": 0.6,
                "merge_adjacent": True,
            },
        },
    ]
    descriptors = _with_descriptor_seeds(raw_descriptors, 777)
    text = "Guard the vault at midnight"
    first = zoo_rust.compose_glitchlings(text, descriptors, 777)
    second = zoo_rust.compose_glitchlings(text, descriptors, 777)
    assert first == second == _run_python_sequence(text, descriptors, 777)


def test_compose_glitchlings_propagates_glitch_errors():
    zoo_rust = pytest.importorskip("glitchlings._zoo_rust")
    master_seed = 404
    descriptors = _with_descriptor_seeds(
        [
            {
                "name": "Redactyl",
                "operation": {
                    "type": "redact",
                    "replacement_char": redactyl_module.FULL_BLOCK,
                    "redaction_rate": 1.0,
                    "merge_adjacent": False,
                },
            }
        ],
        master_seed,
    )
    with pytest.raises(ValueError, match="contains no redactable words"):
        zoo_rust.compose_glitchlings("   \t", descriptors, master_seed)


def test_gaggle_prefers_rust_pipeline(monkeypatch):
    zoo_rust = pytest.importorskip("glitchlings._zoo_rust")
    original_compose = zoo_rust.compose_glitchlings
    calls: list[tuple[str, list[dict[str, object]], int]] = []

    def spy(text: str, descriptors: list[dict[str, object]], master_seed: int) -> str:
        calls.append((text, descriptors, master_seed))
        return original_compose(text, descriptors, master_seed)

    monkeypatch.setenv("GLITCHLINGS_RUST_PIPELINE", "1")
    monkeypatch.setattr(zoo_rust, "compose_glitchlings", spy)
    monkeypatch.setattr(core_module, "_compose_glitchlings_rust", spy, raising=False)

    def _fail(*_args: object, **_kwargs: object) -> str:
        raise AssertionError("Python fallback invoked")

    monkeypatch.setattr(reduple_module, "reduplicate_words", _fail)
    monkeypatch.setattr(rushmore_module, "delete_random_words", _fail)
    monkeypatch.setattr(redactyl_module, "redact_words", _fail)
    monkeypatch.setattr(scannequin_module, "ocr_artifacts", _fail)

    gaggle_glitchlings = [
        scannequin_module.Scannequin(rate=0.2),
        reduple_module.Reduple(rate=0.4),
        rushmore_module.Rushmore(rate=0.3),
        redactyl_module.Redactyl(rate=0.5, merge_adjacent=True),
    ]
    gaggle = core_module.Gaggle(gaggle_glitchlings, seed=777)

    text = "Safeguard the archive tonight"
    result = gaggle(text)
    assert calls, "Expected the Rust pipeline to be invoked"
    descriptors = calls[0][1]
    apply_names = [glitch.name for glitch in gaggle.apply_order]
    original_names = [glitch.name for glitch in gaggle_glitchlings]
    assert apply_names != original_names, "Expected Gaggle to reorder glitchlings"
    expected_seeds = {
        glitch.name: core_module.Gaggle.derive_seed(777, glitch.name, index)
        for index, glitch in enumerate(gaggle_glitchlings)
    }
    assert [descriptor["seed"] for descriptor in descriptors] == [
        expected_seeds[descriptor["name"]]
        for descriptor in descriptors
    ]
    expected = _run_python_sequence(text, descriptors, 777)
    assert result == expected


def test_gaggle_python_fallback_when_pipeline_disabled(monkeypatch):
    pytest.importorskip("glitchlings._zoo_rust")

    def _fail(*_args: object, **_kwargs: object) -> str:
        raise AssertionError("Rust pipeline should not run when feature flag is disabled")

    monkeypatch.delenv("GLITCHLINGS_RUST_PIPELINE", raising=False)
    monkeypatch.setattr(core_module, "_compose_glitchlings_rust", _fail, raising=False)

    gaggle = core_module.Gaggle(
        [
            reduple_module.Reduple(rate=0.4),
            rushmore_module.Rushmore(rate=0.3),
        ],
        seed=2024,
    )

    text = "Hold the door"
    result = gaggle(text)
    raw_descriptors = [
        {"name": "Reduple", "operation": {"type": "reduplicate", "reduplication_rate": 0.4}},
        {"name": "Rushmore", "operation": {"type": "delete", "max_deletion_rate": 0.3}},
    ]
    descriptors = _with_descriptor_seeds(raw_descriptors, 2024)
    expected = _run_python_sequence(text, descriptors, 2024)
    assert result == expected


def test_pipeline_falls_back_for_unsupported_glitchling(monkeypatch):
    master_seed = 1122
    text = "Synchronize thrusters before ascent"

    def _make_glitchlings() -> list[core_module.Glitchling]:
        typo = typogre_module.Typogre(rate=0.02, seed=5)
        redup = reduple_module.Reduple(rate=0.2, seed=7)
        return [typo, redup]

    python_gaggle = core_module.Gaggle(_make_glitchlings(), seed=master_seed)
    expected = python_gaggle(text)

    monkeypatch.setenv("GLITCHLINGS_RUST_PIPELINE", "1")

    def _fail(*_args: object, **_kwargs: object) -> str:
        raise AssertionError("Rust pipeline should not run on unsupported glitchlings")

    monkeypatch.setattr(core_module, "_compose_glitchlings_rust", _fail, raising=False)
    pipeline_gaggle = core_module.Gaggle(_make_glitchlings(), seed=master_seed)

    assert pipeline_gaggle._pipeline_descriptors() is None
    assert pipeline_gaggle(text) == expected


def test_pipeline_falls_back_for_incomplete_operation(monkeypatch):
    master_seed = 909
    text = "Route traffic through the backup relay"

    def _make_glitchlings() -> list[core_module.Glitchling]:
        red = redactyl_module.Redactyl(rate=0.5, merge_adjacent=False, seed=11)
        red.set_param("merge_adjacent", None)
        rush = rushmore_module.Rushmore(rate=0.25, seed=13)
        return [red, rush]

    python_gaggle = core_module.Gaggle(_make_glitchlings(), seed=master_seed)
    expected = python_gaggle(text)

    monkeypatch.setenv("GLITCHLINGS_RUST_PIPELINE", "1")

    def _fail(*_args: object, **_kwargs: object) -> str:
        raise AssertionError("Rust pipeline should not run with incomplete operations")

    monkeypatch.setattr(core_module, "_compose_glitchlings_rust", _fail, raising=False)
    pipeline_gaggle = core_module.Gaggle(_make_glitchlings(), seed=master_seed)

    assert pipeline_gaggle._pipeline_descriptors() is None
    assert pipeline_gaggle(text) == expected


def test_rust_pipeline_feature_flag_introspection(monkeypatch):
    monkeypatch.delenv("GLITCHLINGS_RUST_PIPELINE", raising=False)
    assert not core_module._pipeline_feature_flag_enabled()
    assert core_module.Gaggle.rust_pipeline_supported() is (
        core_module._compose_glitchlings_rust is not None
    )
    assert not core_module.Gaggle.rust_pipeline_enabled()

    monkeypatch.setenv("GLITCHLINGS_RUST_PIPELINE", "1")
    if core_module.Gaggle.rust_pipeline_supported():
        assert core_module.Gaggle.rust_pipeline_enabled()
    else:
        assert not core_module.Gaggle.rust_pipeline_enabled()

    monkeypatch.setenv("GLITCHLINGS_RUST_PIPELINE", "false")
    assert not core_module._pipeline_feature_flag_enabled()
    assert not core_module.Gaggle.rust_pipeline_enabled()






from __future__ import annotations

from glitchlings.zoo.jargoyle import substitute_random_synonyms


def _clean_tokens(text: str) -> list[str]:
    return [token.strip(".,") for token in text.split()]


def test_jargoyle_multiple_pos_targets_change_words():
    text = "They sing happy songs."
    result = substitute_random_synonyms(
        text,
        rate=1.0,
        part_of_speech=("v", "a"),
        seed=123,
    )

    original_tokens = _clean_tokens(text)
    result_tokens = _clean_tokens(result)

    # Expect both verb and adjective replacements to differ from input
    changed = {
        orig for orig, new in zip(original_tokens, result_tokens) if orig != new
    }
    assert {"sing", "happy"} <= changed


def test_jargoyle_any_includes_all_supported_pos():
    text = "They sing happy songs quickly."
    result = substitute_random_synonyms(
        text,
        rate=1.0,
        part_of_speech="any",
        seed=99,
    )

    original_tokens = _clean_tokens(text)
    result_tokens = _clean_tokens(result)

    changed = {
        orig for orig, new in zip(original_tokens, result_tokens) if orig != new
    }
    assert {"sing", "happy", "songs", "quickly"} <= changed



import pytest

from glitchlings import Typogre, summon
from glitchlings.zoo.core import Gaggle


def test_gaggle_determinism(sample_text):
    g1 = summon(["reduple", "mim1c", "typogre", "rushmore", "redactyl"], seed=777)
    out1 = g1(sample_text)
    g2 = summon(["reduple", "mim1c", "typogre", "rushmore", "redactyl"], seed=777)
    out2 = g2(sample_text)
    assert out1 == out2


def test_gaggle_seed_changes_output(sample_text):
    g1 = summon(["reduple", "mim1c", "typogre", "rushmore", "redactyl"], seed=1)
    out1 = g1(sample_text)
    g2 = summon(["reduple", "mim1c", "typogre", "rushmore", "redactyl"], seed=2)
    out2 = g2(sample_text)
    assert out1 != out2


def test_gaggle_ordering_stable(sample_text):
    # When summoned by name, built-ins choose a stable order by scope, then order, then name
    gaggle = summon(["typogre", "mim1c", "reduple", "rushmore"], seed=42)
    expected = ["Reduple", "Rushmore", "Typogre", "Mim1c"]
    assert [member.name for member in gaggle.apply_order] == expected


def test_gaggle_seed_derivation_regression():
    assert Gaggle.derive_seed(151, "Typogre", 0) == 13006513535068165406
    assert Gaggle.derive_seed(151, "Redactyl", 3) == 12503957440331561761


def test_summon_accepts_parameterized_specification():
    gaggle = summon(["Typogre(rate=0.05)"], seed=151)
    assert len(gaggle.apply_order) == 1
    member = gaggle.apply_order[0]
    assert isinstance(member, Typogre)
    assert member.rate == 0.05
    assert member.max_change_rate == 0.05


def test_summon_rejects_positional_parameter_specifications():
    with pytest.raises(ValueError, match="keyword arguments"):
        summon(["Typogre(0.2)"])



from __future__ import annotations

import ast
from typing import Any

from .typogre import Typogre, typogre
from .mim1c import Mim1c, mim1c
from .jargoyle import Jargoyle, jargoyle, dependencies_available as _jargoyle_available
from .reduple import Reduple, reduple
from .rushmore import Rushmore, rushmore
from .redactyl import Redactyl, redactyl
from .scannequin import Scannequin, scannequin
from .core import Glitchling, Gaggle

__all__ = [
    "Typogre",
    "typogre",
    "Mim1c",
    "mim1c",
    "Jargoyle",
    "jargoyle",
    "Reduple",
    "reduple",
    "Rushmore",
    "rushmore",
    "Redactyl",
    "redactyl",
    "Scannequin",
    "scannequin",
    "Glitchling",
    "Gaggle",
    "summon",
    "BUILTIN_GLITCHLINGS",
    "DEFAULT_GLITCHLING_NAMES",
    "parse_glitchling_spec",
]

_HAS_JARGOYLE = _jargoyle_available()

_BUILTIN_GLITCHLING_LIST: list[Glitchling] = [typogre, mim1c]
if _HAS_JARGOYLE:
    _BUILTIN_GLITCHLING_LIST.append(jargoyle)
_BUILTIN_GLITCHLING_LIST.extend([reduple, rushmore, redactyl, scannequin])

BUILTIN_GLITCHLINGS: dict[str, Glitchling] = {
    glitchling.name.lower(): glitchling for glitchling in _BUILTIN_GLITCHLING_LIST
}

_BUILTIN_GLITCHLING_TYPES: dict[str, type[Glitchling]] = {
    typogre.name.lower(): Typogre,
    mim1c.name.lower(): Mim1c,
    reduple.name.lower(): Reduple,
    rushmore.name.lower(): Rushmore,
    redactyl.name.lower(): Redactyl,
    scannequin.name.lower(): Scannequin,
}
if _HAS_JARGOYLE:
    _BUILTIN_GLITCHLING_TYPES[jargoyle.name.lower()] = Jargoyle

DEFAULT_GLITCHLING_NAMES: list[str] = list(BUILTIN_GLITCHLINGS.keys())


def parse_glitchling_spec(specification: str) -> Glitchling:
    """Return a glitchling instance configured according to ``specification``."""

    text = specification.strip()
    if not text:
        raise ValueError("Glitchling specification cannot be empty.")

    if "(" not in text:
        glitchling = BUILTIN_GLITCHLINGS.get(text.lower())
        if glitchling is None:
            raise ValueError(f"Glitchling '{text}' not found.")
        return glitchling

    if not text.endswith(")"):
        raise ValueError(f"Invalid parameter syntax for glitchling '{text}'.")

    name_part, arg_source = text[:-1].split("(", 1)
    name = name_part.strip()
    if not name:
        raise ValueError(f"Invalid glitchling specification '{text}'.")

    lower_name = name.lower()
    glitchling_type = _BUILTIN_GLITCHLING_TYPES.get(lower_name)
    if glitchling_type is None:
        raise ValueError(f"Glitchling '{name}' not found.")

    try:
        call_expr = ast.parse(f"_({arg_source})", mode="eval").body
    except SyntaxError as exc:
        raise ValueError(
            f"Invalid parameter syntax for glitchling '{name}': {exc.msg}"
        ) from exc

    if not isinstance(call_expr, ast.Call) or call_expr.args:
        raise ValueError(
            f"Glitchling '{name}' parameters must be provided as keyword arguments."
        )

    kwargs: dict[str, Any] = {}
    for keyword in call_expr.keywords:
        if keyword.arg is None:
            raise ValueError(
                f"Glitchling '{name}' does not support unpacking arbitrary keyword arguments."
            )
        try:
            kwargs[keyword.arg] = ast.literal_eval(keyword.value)
        except (ValueError, SyntaxError) as exc:
            raise ValueError(
                f"Failed to parse value for parameter '{keyword.arg}' on glitchling '{name}': {exc}"
            ) from exc

    try:
        return glitchling_type(**kwargs)
    except TypeError as exc:
        raise ValueError(f"Failed to instantiate glitchling '{name}': {exc}") from exc


def summon(glitchlings: list[str | Glitchling], seed: int = 151) -> Gaggle:
    """Summon glitchlings by name (using defaults) or instance (to change parameters)."""

    summoned: list[Glitchling] = []
    for entry in glitchlings:
        if isinstance(entry, Glitchling):
            summoned.append(entry)
            continue

        try:
            summoned.append(parse_glitchling_spec(entry))
        except ValueError as exc:
            raise ValueError(str(exc)) from exc

    return Gaggle(summoned, seed=seed)



import string

from glitchlings.util import KEYNEIGHBORS


def test_standard_layouts_cover_alphabet() -> None:
    letters = set(string.ascii_lowercase)
    for layout_name in (
        "QWERTY",
        "DVORAK",
        "COLEMAK",
        "AZERTY",
        "QWERTZ",
        "SPANISH_QWERTY",
        "SWEDISH_QWERTY",
    ):
        layout = getattr(KEYNEIGHBORS, layout_name)
        missing = letters - set(layout)
        assert not missing, f"{layout_name} missing: {sorted(missing)}"


def test_layout_neighbor_expectations() -> None:
    qwerty = getattr(KEYNEIGHBORS, "QWERTY")
    assert {"q", "w", "s", "z"} <= set(qwerty["a"])

    azerty = getattr(KEYNEIGHBORS, "AZERTY")
    assert {"q", "z"} <= set(azerty["a"])

    dvorak = getattr(KEYNEIGHBORS, "DVORAK")
    assert {"a", "e"} <= set(dvorak["o"])

    colemak = getattr(KEYNEIGHBORS, "COLEMAK")
    assert {"j", "n"} <= set(colemak["h"])

    qwertz = getattr(KEYNEIGHBORS, "QWERTZ")
    assert {"t", "u"} <= set(qwertz["z"])

    spanish = getattr(KEYNEIGHBORS, "SPANISH_QWERTY")
    assert {"s", "d", "x"} <= set(spanish["z"])

    swedish = getattr(KEYNEIGHBORS, "SWEDISH_QWERTY")
    assert {"å", "ö"} <= set(swedish["p"])



from collections.abc import Iterable
from random import Random

import pytest

datasets = pytest.importorskip("datasets")
Dataset = datasets.Dataset

from glitchlings.dlc import huggingface as hf_dlc
from glitchlings.zoo.core import AttackWave, Gaggle, Glitchling


def append_rng_token(text: str, *, rng: Random) -> str:
    """Append a deterministic RNG token to the supplied text."""

    return f"{text}-{rng.randint(0, 999)}"


@pytest.fixture(autouse=True)
def ensure_glitch_installed() -> Iterable[None]:
    hf_dlc.install()
    yield


def test_install_is_idempotent() -> None:
    hf_dlc.install()
    assert hasattr(Dataset, "glitch")


def test_module_exports_dataset_with_glitch_method() -> None:
    assert hf_dlc.Dataset is Dataset

    dataset = hf_dlc.Dataset.from_dict({"text": ["alpha"]})
    result = list(dataset.glitch("typogre", column="text"))

    assert len(result) == 1


def test_dataset_glitch_accepts_gaggle() -> None:
    dataset = Dataset.from_dict({"text": ["alpha", "beta"], "label": [0, 1]})
    glitchling = Glitchling("rngster", append_rng_token, AttackWave.SENTENCE, seed=1337)
    gaggle = Gaggle([glitchling], seed=99)

    corrupted = dataset.glitch(gaggle, column="text")

    comparison_gaggle = Gaggle([glitchling.clone()], seed=99)
    expected = list(comparison_gaggle.corrupt_dataset(dataset, ["text"]))
    assert list(corrupted) == expected
    assert list(dataset)[0]["text"] == "alpha"


def test_dataset_glitch_accepts_names_and_respects_seed() -> None:
    dataset = Dataset.from_dict({"text": ["alpha", "beta"]})

    corrupted = list(dataset.glitch("typogre", column="text", seed=42))
    rerun = list(dataset.glitch(["Typogre"], column="text", seed=42))

    assert corrupted == rerun



import inspect
from unittest.mock import patch

from glitchlings.zoo.core import AttackWave, Glitchling
from glitchlings.zoo.typogre import Typogre


def test_typogre_clone_preserves_configuration_and_seed_behavior() -> None:
    original = Typogre(rate=0.05, keyboard="AZERTY", seed=111)

    clone = original.clone(seed=222)

    assert isinstance(clone, Typogre)
    assert clone.rate == original.rate
    assert clone.max_change_rate == original.rate
    assert clone.keyboard == original.keyboard

    sample_text = "The quick brown fox jumps over the lazy dog."

    original.reset_rng()
    original_result = original(sample_text)

    clone.reset_rng()
    clone_result_first = clone(sample_text)
    clone.reset_rng()
    clone_result_second = clone(sample_text)

    assert clone_result_first == clone_result_second
    assert clone_result_first != original_result

def test_glitchling_signature_introspection_is_cached() -> None:
    call_count = 0
    original_signature = inspect.signature

    def tracking_signature(func: object) -> inspect.Signature:
        nonlocal call_count
        call_count += 1
        return original_signature(func)

    def corruption(text: str, *, rng: object) -> str:
        assert rng is not None
        return text.upper()

    glitchling = Glitchling("CacheTester", corruption, AttackWave.DOCUMENT)

    with patch("glitchlings.zoo.core.inspect.signature", side_effect=tracking_signature):
        glitchling.corrupt("hello")
        glitchling.corrupt("world")

    assert call_count == 1

def test_glitchling_pipeline_operation_factory_survives_clone() -> None:
    def descriptor(glitchling: Glitchling) -> dict[str, object]:
        return {"type": "custom", "value": glitchling.kwargs.get("value")}

    glitch = Glitchling(
        "Factory",
        lambda text, **_: text,
        AttackWave.WORD,
        pipeline_operation=descriptor,
        value=7,
    )

    assert glitch.pipeline_operation() == {"type": "custom", "value": 7}

    clone = glitch.clone()
    assert clone.pipeline_operation() == {"type": "custom", "value": 7}




from __future__ import annotations

import sys
import types

import pytest


class _Rubric:
    def __init__(self, funcs, weights):
        self.funcs = list(funcs)
        self.weights = list(weights)


class _SingleTurnEnv:
    def __init__(self, dataset=None, rubric=None):
        self.dataset = dataset
        self.rubric = rubric

class _VerifierEnvironment:
    def __init__(self, dataset=None):
        self.dataset = dataset


def _load_environment(_: str) -> _VerifierEnvironment:
    return _VerifierEnvironment()


verifiers_stub = types.ModuleType("verifiers")
verifiers_stub.Environment = _VerifierEnvironment
verifiers_stub.Rubric = _Rubric
verifiers_stub.SingleTurnEnv = _SingleTurnEnv
verifiers_stub.load_environment = _load_environment
sys.modules["verifiers"] = verifiers_stub

import glitchlings.zoo.core as zoo_core
import glitchlings.zoo.jargoyle as jargoyle
from glitchlings.zoo.core import AttackWave, Gaggle, Glitchling
from glitchlings.dlc import prime


def _no_op_ensure_wordnet() -> None:
    jargoyle._wordnet_ready = True


jargoyle.ensure_wordnet = _no_op_ensure_wordnet
jargoyle._ensure_wordnet = _no_op_ensure_wordnet
jargoyle._wordnet_ready = True




class FakeDataset:
    def __init__(self, rows: list[dict[str, object]], column_names: list[str] | None = None, *, streaming: bool = False):
        self._rows = [dict(row) for row in rows]
        if column_names is None:
            if rows:
                column_names = list(rows[0].keys())
            else:
                column_names = []
        self.column_names = list(column_names)
        self._streaming = streaming

    @classmethod
    def from_dict(cls, columns: dict[str, list[object]], *, streaming: bool = False) -> "FakeDataset":
        keys = list(columns.keys())
        lengths = [len(col) for col in columns.values()]
        if lengths and any(l != lengths[0] for l in lengths):
            raise ValueError(f"All columns must have the same length, but got lengths: {dict(zip(keys, lengths))}")
        length = lengths[0] if lengths else 0
        rows = [
            {key: columns[key][index] for key in keys}
            for index in range(length)
        ]
        return cls(rows, keys, streaming=streaming)

    def __len__(self) -> int:
        if self._streaming:
            raise TypeError("Streaming dataset does not define __len__.")
        return len(self._rows)

    def __getitem__(self, index: int) -> dict[str, object]:
        return dict(self._rows[index])

    def __iter__(self):
        for row in self._rows:
            yield dict(row)

    def filter(self, function, load_from_cache_file: bool = True):
        filtered = [row for row in self._rows if function(dict(row))]
        return FakeDataset(filtered, self.column_names, streaming=self._streaming)

    def map(
        self,
        function,
        remove_columns=None,
        load_from_cache_file: bool = True,
    ):
        mapped_rows = []
        for row in self._rows:
            result = function(dict(row))
            if remove_columns:
                for column in remove_columns:
                    result.pop(column, None)
            mapped_rows.append(result)
        if mapped_rows:
            column_names = list(mapped_rows[0].keys())
        else:
            column_names = []
        return FakeDataset(mapped_rows, column_names, streaming=self._streaming)

    def take(self, n: int):
        return FakeDataset(self._rows[:n], self.column_names, streaming=self._streaming)

    def with_transform(self, function):
        transformed = [function(dict(row)) for row in self._rows]
        return FakeDataset(transformed, self.column_names, streaming=self._streaming)


Dataset = FakeDataset


@pytest.fixture(autouse=True)
def _install_fake_datasets(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(zoo_core, "_DatasetsDataset", FakeDataset, raising=True)
    monkeypatch.setattr(zoo_core, "_datasets_error", None, raising=True)


def append_marker(text: str) -> str:
    """Tag the provided text with a deterministic marker."""

    return f"{text}<<<"


def test_conversational_prompts_remain_structured() -> None:
    dataset = Dataset.from_dict(
        {
            "prompt": [
                [
                    {"role": "system", "content": "Restore the text."},
                    {"role": "user", "content": "coRRuPt3d"},
                ]
            ]
        }
    )

    glitchling = Glitchling("marker", append_marker, AttackWave.SENTENCE)
    gaggle = Gaggle([glitchling], seed=99)

    corrupted_rows = list(gaggle.corrupt_dataset(dataset, ["prompt"]))

    assert len(corrupted_rows) == 1
    prompt = corrupted_rows[0]["prompt"]

    assert isinstance(prompt, list)
    assert prompt[0] == {"role": "system", "content": "Restore the text."}
    assert prompt[1]["role"] == "user"
    assert prompt[1]["content"] == "coRRuPt3d<<<"


def test_prime_resolve_columns_requires_string_candidates():
    dataset = Dataset.from_dict({"scores": [[1, 2], [3, 4]], "ids": [1, 2]})

    with pytest.raises(ValueError, match=r"Unable to determine which dataset columns to corrupt\."):
        prime._resolve_columns(dataset, None)


class _FakeEnvironment:
    def __init__(self, dataset):
        self.dataset = dataset


class _RecordingGaggle:
    def __init__(self):
        self.columns_seen: list[list[str]] = []

    def corrupt_dataset(self, dataset, columns):
        self.columns_seen.append(list(columns))
        return dataset


def test_load_environment_respects_explicit_columns(monkeypatch):
    dataset = Dataset.from_dict({"prompt": ["alpha"], "extra": ["beta"]})
    stub = _RecordingGaggle()

    monkeypatch.setattr(prime, "_resolve_environment", lambda _env: _FakeEnvironment(dataset))
    monkeypatch.setattr(prime, "summon", lambda specs, seed: stub)

    env = prime.load_environment("ignored", glitchlings=[prime.Typogre()], seed=7, columns=["extra"])

    assert env.dataset is dataset
    assert stub.columns_seen == [["extra"]]


def test_tutorial_level_applies_tuned_glitchlings(monkeypatch):
    baseline = "alpha beta gamma delta"

    def _fake_environment(_):
        return _FakeEnvironment(Dataset.from_dict({"prompt": [baseline]}))

    monkeypatch.setattr(prime, "_resolve_environment", _fake_environment)

    env = prime.tutorial_level("ignored", seed=123, difficulty=prime.Difficulty.Easy)
    mutated_prompt = list(env.dataset)[0]["prompt"]
    assert mutated_prompt != baseline

    env_again = prime.tutorial_level("ignored", seed=123, difficulty=prime.Difficulty.Easy)
    mutated_again = list(env_again.dataset)[0]["prompt"]
    assert mutated_again != baseline
    assert mutated_prompt == mutated_again


def test_extract_completion_text_handles_strings_and_structures() -> None:
    assert prime._extract_completion_text("plain text") == "plain text"

    message = [{"role": "assistant", "content": "structured"}]
    assert prime._extract_completion_text(message) == "structured"

    assert prime._extract_completion_text({"foo": 1}) == "{'foo': 1}"


def test_similarity_scores_expected_fraction_for_single_edit() -> None:
    completion = "hello"
    answer = "hella"
    expected_distance = 1
    expected_score = 1 - (expected_distance / max(len(completion), len(answer), 1))

    score = prime.symmetric_damerau_levenshtein_similarity(None, completion, answer)

    assert score == pytest.approx(expected_score)


def test_similarity_handles_identical_and_extreme_inputs() -> None:
    assert prime.symmetric_damerau_levenshtein_similarity(None, "same", "same") == 1.0

    shorter = prime.symmetric_damerau_levenshtein_similarity(None, "a", "a" * 10)
    longer = prime.symmetric_damerau_levenshtein_similarity(None, "a" * 10, "a")
    assert 0.0 <= shorter <= 1.0
    assert 0.0 <= longer <= 1.0

    empty_answer = prime.symmetric_damerau_levenshtein_similarity(None, "abc", "")
    both_empty = prime.symmetric_damerau_levenshtein_similarity(None, "", "")
    assert empty_answer == 0.0
    assert both_empty == 1.0
def test_echo_chamber_streams_dataset(monkeypatch: pytest.MonkeyPatch) -> None:
    base_dataset = Dataset.from_dict({
        "text": ["alpha", None, "beta"],
        "other": [1, 2, 3],
    }, streaming=True)

    def _fake_load_dataset(*args, **kwargs):
        return base_dataset

    datasets_stub = types.ModuleType("datasets")
    datasets_stub.Dataset = Dataset
    datasets_stub.DatasetDict = dict
    datasets_stub.load_dataset = _fake_load_dataset
    monkeypatch.setitem(sys.modules, "datasets", datasets_stub)

    class _RecordingGaggle:
        def __init__(self):
            self.records: list[tuple[Dataset, list[str]]] = []

        def corrupt_dataset(self, dataset, columns):
            self.records.append((dataset, list(columns)))
            return dataset

    recorder = _RecordingGaggle()
    monkeypatch.setattr(prime, "_as_gaggle", lambda glitchlings, seed: recorder)

    env = prime.echo_chamber(
        dataset_id="stub/the-dataset",
        column="text",
        glitchlings=["Typogre"],
        instructions="Restore the text.",
    )

    assert sum(1 for _ in base_dataset) == 3
    assert recorder.records
    dataset, columns = recorder.records[0]
    assert dataset is not base_dataset
    assert columns == ["prompt"]

    assert env.dataset is dataset
    rows = list(dataset)
    assert len(rows) == 2
    assert rows[0]["answer"] == "alpha"
    assert rows[0]["prompt"][0]["content"] == "Restore the text."
    assert rows[0]["prompt"][1]["content"] == "Corrupted text:\nalpha"








import re
import random
from typing import Any

from .core import Glitchling, AttackWave
from ._rate import resolve_rate

FULL_BLOCK = "█"


try:
    from glitchlings._zoo_rust import redact_words as _redact_words_rust
except ImportError:  # pragma: no cover - compiled extension not present
    _redact_words_rust = None


def _python_redact_words(
    text: str,
    *,
    replacement_char: str,
    rate: float,
    merge_adjacent: bool,
    rng: random.Random,
) -> str:
    """Redact random words by replacing their characters.

    Parameters
    - text: Input text.
    - replacement_char: The character to use for redaction (default FULL_BLOCK).
    - rate: Max proportion of words to redact (default 0.05).
    - merge_adjacent: If True, merges adjacent redactions across intervening non-word chars.
    - seed: Seed used if `rng` not provided (default 151).
    - rng: Optional RNG; overrides seed.
    """
    # Preserve exact spacing and punctuation by using regex
    tokens = re.split(r"(\s+)", text)
    word_indices = [i for i, token in enumerate(tokens) if i % 2 == 0 and token.strip()]
    if not word_indices:
        raise ValueError("Cannot redact words because the input text contains no redactable words.")
    num_to_redact = max(1, int(len(word_indices) * rate))

    # Sample from the indices of actual words
    indices_to_redact = rng.sample(word_indices, k=num_to_redact)
    indices_to_redact.sort()

    for i in indices_to_redact:
        if i >= len(tokens):
            break

        word = tokens[i]
        if not word or word.isspace():  # Skip empty or whitespace
            continue

        # Check if word has trailing punctuation
        match = re.match(r"^(\W*)(.*?)(\W*)$", word)
        if match:
            prefix, core, suffix = match.groups()
            tokens[i] = f"{prefix}{replacement_char * len(core)}{suffix}"
        else:
            tokens[i] = f"{replacement_char * len(word)}"

    text = "".join(tokens)

    if merge_adjacent:
        text = re.sub(
            rf"{replacement_char}\W+{replacement_char}",
            lambda m: replacement_char * (len(m.group(0)) - 1),
            text,
        )

    return text


def redact_words(
    text: str,
    replacement_char: str = FULL_BLOCK,
    rate: float | None = None,
    merge_adjacent: bool = False,
    seed: int = 151,
    rng: random.Random | None = None,
    *,
    redaction_rate: float | None = None,
) -> str:
    """Redact random words by replacing their characters."""

    effective_rate = resolve_rate(
        rate=rate,
        legacy_value=redaction_rate,
        default=0.05,
        legacy_name="redaction_rate",
    )

    if rng is None:
        rng = random.Random(seed)

    clamped_rate = max(0.0, effective_rate)

    use_rust = _redact_words_rust is not None and isinstance(merge_adjacent, bool)

    if use_rust:
        return _redact_words_rust(
            text,
            replacement_char,
            clamped_rate,
            merge_adjacent,
            rng,
        )

    return _python_redact_words(
        text,
        replacement_char=replacement_char,
        rate=clamped_rate,
        merge_adjacent=merge_adjacent,
        rng=rng,
    )


class Redactyl(Glitchling):
    """Glitchling that redacts words with block characters."""

    def __init__(
        self,
        *,
        replacement_char: str = FULL_BLOCK,
        rate: float | None = None,
        redaction_rate: float | None = None,
        merge_adjacent: bool = False,
        seed: int = 151,
    ) -> None:
        self._param_aliases = {"redaction_rate": "rate"}
        effective_rate = resolve_rate(
            rate=rate,
            legacy_value=redaction_rate,
            default=0.05,
            legacy_name="redaction_rate",
        )
        super().__init__(
            name="Redactyl",
            corruption_function=redact_words,
            scope=AttackWave.WORD,
            seed=seed,
            replacement_char=replacement_char,
            rate=effective_rate,
            merge_adjacent=merge_adjacent,
        )

    def pipeline_operation(self) -> dict[str, Any] | None:
        replacement_char = self.kwargs.get("replacement_char")
        rate = self.kwargs.get("rate")
        merge_adjacent = self.kwargs.get("merge_adjacent")
        if replacement_char is None or rate is None or merge_adjacent is None:
            return None
        return {
            "type": "redact",
            "replacement_char": str(replacement_char),
            "redaction_rate": float(rate),
            "merge_adjacent": bool(merge_adjacent),
        }



redactyl = Redactyl()


__all__ = ["Redactyl", "redactyl"]



"""Regression tests covering dataset corruption with optional dependencies."""

from __future__ import annotations

import builtins
import importlib
import random
import sys
from typing import Any

import pytest


def _materialize(dataset: "Dataset") -> list[dict[str, Any]]:
    """Eagerly load a dataset to regular dictionaries for comparison."""

    return [
        {column: row[column] for column in dataset.column_names}
        for row in dataset
    ]


def test_corrupt_dataset_is_deterministic_across_columns() -> None:
    """Ensure ``Glitchling.corrupt_dataset`` produces stable results."""

    pytest.importorskip("datasets")
    from datasets import Dataset

    from glitchlings.zoo.core import AttackWave, Glitchling

    def append_rng_token(text: str, *, rng: random.Random) -> str:
        return f"{text}:{rng.randint(0, 999)}"

    dataset = Dataset.from_dict(
        {
            "text": ["alpha", "bravo", "charlie"],
            "untouched": ["keep", "these", "same"],
        }
    )
    glitchling = Glitchling(
        "rngster", append_rng_token, AttackWave.SENTENCE, seed=2024
    )

    original_rows = _materialize(dataset)

    glitchling.reset_rng(seed=1337)
    corrupted_first = glitchling.corrupt_dataset(dataset, columns=["text"])
    first_rows = _materialize(corrupted_first)

    glitchling.reset_rng(seed=1337)
    corrupted_second = glitchling.corrupt_dataset(dataset, columns=["text"])
    second_rows = _materialize(corrupted_second)

    assert corrupted_first.column_names == dataset.column_names
    assert corrupted_second.column_names == dataset.column_names
    assert first_rows == second_rows
    assert [row["untouched"] for row in first_rows] == [
        row["untouched"] for row in original_rows
    ]


def test_corrupt_dataset_requires_optional_dependency(monkeypatch: pytest.MonkeyPatch) -> None:
    """Ensure ``Glitchling.corrupt_dataset`` surfaces a clear error."""

    monkeypatch.delitem(sys.modules, "datasets", raising=False)

    original_import = builtins.__import__

    def fake_import(name: str, *args: object, **kwargs: object):  # type: ignore[override]
        if name == "datasets":
            raise ModuleNotFoundError("datasets is not installed")
        return original_import(name, *args, **kwargs)

    monkeypatch.setattr(builtins, "__import__", fake_import)

    for module_name in [
        module
        for module in sys.modules
        if module == "glitchlings" or module.startswith("glitchlings.")
    ]:
        monkeypatch.delitem(sys.modules, module_name, raising=False)

    glitchlings = importlib.import_module("glitchlings")
    assert glitchlings is not None

    from glitchlings.zoo.core import AttackWave, Glitchling

    noop = Glitchling("noop", lambda text, **_: text, AttackWave.CHARACTER)

    with pytest.raises(ModuleNotFoundError, match="datasets is not installed"):
        noop.corrupt_dataset(dataset=object(), columns=["text"])


def test_corrupt_dataset_glitches_list_columns_individually():
    pytest.importorskip("datasets")
    from datasets import Dataset

    from glitchlings.zoo.core import AttackWave, Glitchling

    def tag_entry(text: str, *, rng: random.Random) -> str:
        return f"{text}:{rng.randint(0, 99)}"

    dataset = Dataset.from_dict(
        {
            "utterances": [["alpha", "bravo"], ["charlie", "delta"]],
            "notes": ["keep", "steady"],
        }
    )
    glitchling = Glitchling("tagger", tag_entry, AttackWave.SENTENCE, seed=314)

    glitchling.reset_rng(seed=21)
    corrupted = glitchling.corrupt_dataset(dataset, columns=["utterances"])

    rows = _materialize(corrupted)
    glitchling.reset_rng(seed=21)
    repeat_rows = _materialize(glitchling.corrupt_dataset(dataset, columns=["utterances"]))

    assert rows == repeat_rows
    originals = dataset["utterances"]
    original_notes = dataset["notes"]
    for row, original, note in zip(rows, originals, original_notes):
        mutated = row["utterances"]
        assert isinstance(mutated, str)
        for item in original:
            assert item in mutated
        assert ":" in mutated
        assert row["notes"] == note




from functools import partial
from typing import cast

from glitchlings import (
    typogre,
    mim1c,
    jargoyle,
    reduple,
    rushmore,
    redactyl,
    scannequin,
)
from glitchlings.zoo.core import AttackWave, Glitchling


def _twice(fn, text: str, seed: int = 42) -> tuple[str, str]:
    fn.reset_rng(seed)
    out1 = cast(str, fn(text))
    fn.reset_rng(seed)
    out2 = cast(str, fn(text))
    return out1, out2


def test_typogre_determinism(sample_text):
    typogre.set_param("seed", 42)
    typogre.set_param("rate", 0.03)
    a, b = _twice(typogre, sample_text)
    assert a == b


def test_mim1c_determinism(sample_text):
    mim1c.set_param("seed", 42)
    mim1c.set_param("rate", 0.03)
    mim1c.set_param("classes", ["LATIN", "GREEK", "CYRILLIC"])  # explicit default
    a, b = _twice(mim1c, sample_text)
    assert a == b


def test_jargoyle_determinism(sample_text):
    jargoyle.set_param("seed", 42)
    jargoyle.set_param("rate", 0.05)
    a, b = _twice(jargoyle, sample_text)
    assert a == b


def test_reduple_determinism(sample_text):
    reduple.set_param("seed", 42)
    reduple.set_param("rate", 0.05)
    a, b = _twice(reduple, sample_text)
    assert a == b


def test_rushmore_determinism(sample_text):
    rushmore.set_param("seed", 42)
    rushmore.set_param("rate", 0.01)
    a, b = _twice(rushmore, sample_text)
    assert a == b


def test_redactyl_determinism(sample_text):
    redactyl.set_param("seed", 42)
    redactyl.set_param("rate", 0.05)
    a, b = _twice(redactyl, sample_text)
    assert a == b


def test_scannequin_determinism(sample_text):
    scannequin.set_param("seed", 42)
    scannequin.set_param("rate", 0.03)
    a, b = _twice(scannequin, sample_text)
    assert a == b


def _partial_helper(text: str, *, rng) -> str:
    # The helper intentionally relies on the injected RNG to ensure it is provided.
    rng.random()
    return text


def test_partial_corruption_receives_rng(sample_text):
    """Ensure RNG injection works for callables lacking a ``__code__`` attribute."""

    corruption = partial(_partial_helper)
    glitchling = Glitchling("partial", corruption, AttackWave.WORD)

    result = glitchling(sample_text)
    assert result == sample_text



from .zoo import (
    Typogre,
    typogre,
    Mim1c,
    mim1c,
    Jargoyle,
    jargoyle,
    Redactyl,
    redactyl,
    Reduple,
    reduple,
    Rushmore,
    rushmore,
    Scannequin,
    scannequin,
    Glitchling,
    Gaggle,
    summon,
)
from .util import SAMPLE_TEXT


__all__ = [
    "Typogre",
    "typogre",
    "Mim1c",
    "mim1c",
    "Jargoyle",
    "jargoyle",
    "Redactyl",
    "redactyl",
    "Reduple",
    "reduple",
    "Rushmore",
    "rushmore",
    "Scannequin",
    "scannequin",
    "summon",
    "Glitchling",
    "Gaggle",
    "SAMPLE_TEXT",
]



"""Integration helpers for the Hugging Face datasets library."""

from __future__ import annotations

from collections.abc import Iterable, Sequence
from typing import Any

try:  # pragma: no cover - optional dependency is required at runtime
    from datasets import Dataset as _DatasetsDataset
except ModuleNotFoundError as _datasets_error:  # pragma: no cover - optional dependency
    _DatasetsDataset = None  # type: ignore[assignment]
else:
    _datasets_error = None

from ..zoo import Gaggle, Glitchling, summon


def _normalise_columns(column: str | Sequence[str]) -> list[str]:
    """Normalise a column specification to a list."""

    if isinstance(column, str):
        return [column]

    normalised = list(column)
    if not normalised:
        raise ValueError("At least one column must be specified")
    return normalised


def _as_gaggle(glitchlings: Glitchling | Gaggle | str | Iterable[str | Glitchling], seed: int) -> Gaggle:
    """Coerce any supported glitchling specification into a :class:`Gaggle`."""

    if isinstance(glitchlings, Gaggle):
        return glitchlings

    if isinstance(glitchlings, (Glitchling, str)):
        resolved: Iterable[str | Glitchling] = [glitchlings]
    else:
        resolved = glitchlings

    return summon(list(resolved), seed=seed)


def _glitch_dataset(
    dataset: Any,
    glitchlings: Glitchling | Gaggle | str | Iterable[str | Glitchling],
    column: str | Sequence[str],
    *,
    seed: int = 151,
) -> Any:
    """Internal helper implementing :meth:`Dataset.glitch`."""

    columns = _normalise_columns(column)
    gaggle = _as_gaggle(glitchlings, seed=seed)
    return gaggle.corrupt_dataset(dataset, columns)


def _ensure_dataset_class() -> Any:
    """Return the Hugging Face :class:`~datasets.Dataset` patched with ``.glitch``."""

    if _DatasetsDataset is None:  # pragma: no cover - datasets is an install-time dependency
        message = "datasets is not installed"
        raise ModuleNotFoundError(message) from _datasets_error

    if getattr(_DatasetsDataset, "glitch", None) is None:

        def glitch(  # type: ignore[override]
            self: Any,
            glitchlings: Glitchling | Gaggle | str | Iterable[str | Glitchling],
            *,
            column: str | Sequence[str],
            seed: int = 151,
            **_: Any,
        ) -> Any:
            """Return a lazily corrupted copy of the dataset."""

            return _glitch_dataset(self, glitchlings, column, seed=seed)

        setattr(_DatasetsDataset, "glitch", glitch)

    return _DatasetsDataset


def install() -> None:
    """Monkeypatch the Hugging Face :class:`~datasets.Dataset` with ``.glitch``."""

    _ensure_dataset_class()


if _DatasetsDataset is not None:
    Dataset = _ensure_dataset_class()
else:  # pragma: no cover - datasets is an install-time dependency
    Dataset = None  # type: ignore[assignment]


__all__ = ["Dataset", "install"]



import math
import random
import re
from typing import Any

from .core import Glitchling, AttackWave
from ._rate import resolve_rate

try:
    from glitchlings._zoo_rust import delete_random_words as _delete_random_words_rust
except ImportError:  # pragma: no cover - compiled extension not present
    _delete_random_words_rust = None


def _python_delete_random_words(
    text: str,
    *,
    rate: float,
    rng: random.Random,
) -> str:
    """Delete random words from the input text while preserving whitespace."""

    if rate <= 0.0:
        return text

    tokens = re.split(r"(\s+)", text)  # Split but keep separators for later rejoin

    candidate_indices: list[int] = []
    for i in range(2, len(tokens), 2):  # Every other token is a word, skip the first word
        word = tokens[i]
        if not word or word.isspace():
            continue

        candidate_indices.append(i)

    allowed_deletions = min(
        len(candidate_indices), math.floor(len(candidate_indices) * rate)
    )
    if allowed_deletions <= 0:
        return text

    deletions = 0
    for i in candidate_indices:
        if rng.random() < rate:
            word = tokens[i]
            match = re.match(r"^(\W*)(.*?)(\W*)$", word)
            if match:
                prefix, _, suffix = match.groups()
                tokens[i] = f"{prefix.strip()}{suffix.strip()}"
            else:
                tokens[i] = ""

            deletions += 1
            if deletions >= allowed_deletions:
                break

    text = "".join(tokens)
    text = re.sub(r"\s+([.,;:])", r"\1", text)
    text = re.sub(r"\s{2,}", " ", text).strip()

    return text


def delete_random_words(
    text: str,
    rate: float | None = None,
    seed: int | None = None,
    rng: random.Random | None = None,
    *,
    max_deletion_rate: float | None = None,
) -> str:
    """Delete random words from the input text.

    Uses the optional Rust implementation when available.
    """

    effective_rate = resolve_rate(
        rate=rate,
        legacy_value=max_deletion_rate,
        default=0.01,
        legacy_name="max_deletion_rate",
    )

    if rng is None:
        rng = random.Random(seed)

    clamped_rate = max(0.0, effective_rate)

    if _delete_random_words_rust is not None:
        return _delete_random_words_rust(text, clamped_rate, rng)

    return _python_delete_random_words(
        text,
        rate=clamped_rate,
        rng=rng,
    )


class Rushmore(Glitchling):
    """Glitchling that deletes words to simulate missing information."""

    def __init__(
        self,
        *,
        rate: float | None = None,
        max_deletion_rate: float | None = None,
        seed: int | None = None,
    ) -> None:
        self._param_aliases = {"max_deletion_rate": "rate"}
        effective_rate = resolve_rate(
            rate=rate,
            legacy_value=max_deletion_rate,
            default=0.01,
            legacy_name="max_deletion_rate",
        )
        super().__init__(
            name="Rushmore",
            corruption_function=delete_random_words,
            scope=AttackWave.WORD,
            seed=seed,
            rate=effective_rate,
        )

    def pipeline_operation(self) -> dict[str, Any] | None:
        rate = self.kwargs.get("rate")
        if rate is None:
            rate = self.kwargs.get("max_deletion_rate")
        if rate is None:
            return None
        return {"type": "delete", "max_deletion_rate": float(rate)}


rushmore = Rushmore()


__all__ = ["Rushmore", "rushmore"]



import random
import re
from collections.abc import Iterable
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Literal, cast

try:  # pragma: no cover - exercised in environments with NLTK installed
    import nltk  # type: ignore[import]
except ModuleNotFoundError as exc:  # pragma: no cover - triggered when NLTK missing
    nltk = None  # type: ignore[assignment]
    find = None  # type: ignore[assignment]
    _NLTK_IMPORT_ERROR = exc
else:  # pragma: no cover - executed when NLTK is available
    from nltk.corpus.reader import WordNetCorpusReader as _WordNetCorpusReader  # type: ignore[import]
    from nltk.data import find as _nltk_find  # type: ignore[import]

    find = _nltk_find
    _NLTK_IMPORT_ERROR = None

if TYPE_CHECKING:  # pragma: no cover - typing aid only
    from nltk.corpus.reader import WordNetCorpusReader  # type: ignore[import]
else:  # Use ``Any`` at runtime to avoid hard dependency when NLTK missing
    WordNetCorpusReader = Any

if nltk is not None:  # pragma: no cover - guarded by import success
    try:
        from nltk.corpus import wordnet as _WORDNET_MODULE  # type: ignore[import]
    except ModuleNotFoundError:  # pragma: no cover - only hit on namespace packages
        _WORDNET_MODULE = None
    else:
        WordNetCorpusReader = _WordNetCorpusReader  # type: ignore[assignment]
else:
    _WORDNET_MODULE = None

from .core import AttackWave, Glitchling
from ._rate import resolve_rate

_WORDNET_HANDLE: WordNetCorpusReader | Any | None = _WORDNET_MODULE

_wordnet_ready = False


def _require_nltk() -> None:
    """Ensure the NLTK dependency is present before continuing."""

    if nltk is None or find is None:
        message = (
            "The NLTK package is required for the jargoyle glitchling; install "
            "the 'wordnet' extra via `pip install glitchlings[wordnet]`."
        )
        if '_NLTK_IMPORT_ERROR' in globals() and _NLTK_IMPORT_ERROR is not None:
            raise RuntimeError(message) from _NLTK_IMPORT_ERROR
        raise RuntimeError(message)


def dependencies_available() -> bool:
    """Return ``True`` when the runtime NLTK dependency is present."""

    return nltk is not None and find is not None


def _load_wordnet_reader() -> WordNetCorpusReader:
    """Return a WordNet corpus reader from the downloaded corpus files."""

    _require_nltk()

    try:
        root = find("corpora/wordnet")
    except LookupError:
        try:
            zip_root = find("corpora/wordnet.zip")
        except LookupError as exc:
            raise RuntimeError(
                "The NLTK WordNet corpus is not installed; run `nltk.download('wordnet')`."
            ) from exc
        root = zip_root.join("wordnet/")

    return WordNetCorpusReader(root, None)


def _wordnet(force_refresh: bool = False) -> WordNetCorpusReader | Any:
    """Retrieve the active WordNet handle, rebuilding it on demand."""

    global _WORDNET_HANDLE

    if force_refresh:
        _WORDNET_HANDLE = _WORDNET_MODULE

    if _WORDNET_HANDLE is not None:
        return _WORDNET_HANDLE

    _WORDNET_HANDLE = _load_wordnet_reader()
    return _WORDNET_HANDLE


def ensure_wordnet() -> None:
    """Ensure the WordNet corpus is available before use."""

    global _wordnet_ready
    if _wordnet_ready:
        return

    _require_nltk()

    resource = _wordnet()

    try:
        resource.ensure_loaded()
    except LookupError:
        nltk.download("wordnet", quiet=True)
        try:
            resource = _wordnet(force_refresh=True)
            resource.ensure_loaded()
        except LookupError as exc:  # pragma: no cover - only triggered when download fails
            raise RuntimeError(
                "Unable to load NLTK WordNet corpus for the jargoyle glitchling."
            ) from exc

    _wordnet_ready = True


# Backwards compatibility for callers relying on the previous private helper name.
_ensure_wordnet = ensure_wordnet


PartOfSpeech = Literal["n", "v", "a", "r"]
PartOfSpeechInput = PartOfSpeech | Iterable[PartOfSpeech] | Literal["any"]
NormalizedPartsOfSpeech = tuple[PartOfSpeech, ...]

_VALID_POS: tuple[PartOfSpeech, ...] = ("n", "v", "a", "r")


def _split_token(token: str) -> tuple[str, str, str]:
    """Split a token into leading punctuation, core word, and trailing punctuation."""

    match = re.match(r"^(\W*)(.*?)(\W*)$", token)
    if not match:
        return "", token, ""
    prefix, core, suffix = match.groups()
    return prefix, core, suffix


def _normalize_parts_of_speech(part_of_speech: PartOfSpeechInput) -> NormalizedPartsOfSpeech:
    """Coerce user input into a tuple of valid WordNet POS tags."""

    if isinstance(part_of_speech, str):
        lowered = part_of_speech.lower()
        if lowered == "any":
            return _VALID_POS
        if lowered not in _VALID_POS:
            raise ValueError(
                "part_of_speech must be one of 'n', 'v', 'a', 'r', or 'any'"
            )
        return (cast(PartOfSpeech, lowered),)

    normalized: list[PartOfSpeech] = []
    for pos in part_of_speech:
        if pos not in _VALID_POS:
            raise ValueError(
                "part_of_speech entries must be one of 'n', 'v', 'a', or 'r'"
            )
        if pos not in normalized:
            normalized.append(pos)
    if not normalized:
        raise ValueError("part_of_speech iterable may not be empty")
    return tuple(normalized)


@dataclass(frozen=True)
class CandidateInfo:
    """Metadata for a candidate token that may be replaced."""

    prefix: str
    core_word: str
    suffix: str
    parts_of_speech: NormalizedPartsOfSpeech


def _collect_synonyms(
    word: str, parts_of_speech: NormalizedPartsOfSpeech
) -> list[str]:
    """Gather deterministic synonym candidates for the supplied word."""

    normalized_word = word.lower()
    wordnet = _wordnet()
    synonyms: set[str] = set()
    for pos_tag in parts_of_speech:
        synsets = wordnet.synsets(word, pos=pos_tag)
        if not synsets:
            continue

        for synset in synsets:
            lemmas_list = [lemma.name() for lemma in cast(Any, synset).lemmas()]
            if not lemmas_list:
                continue

            filtered = []
            for lemma_str in lemmas_list:
                cleaned = lemma_str.replace("_", " ")
                if cleaned.lower() != normalized_word:
                    filtered.append(cleaned)

            if filtered:
                synonyms.update(filtered)
                break

        if synonyms:
            break

    return sorted(synonyms)


def substitute_random_synonyms(
    text: str,
    rate: float | None = None,
    part_of_speech: PartOfSpeechInput = "n",
    seed: int | None = None,
    rng: random.Random | None = None,
    *,
    replacement_rate: float | None = None,
) -> str:
    """Replace words with random WordNet synonyms.

    Parameters
    - text: Input text.
    - rate: Max proportion of candidate words to replace (default 0.1).
    - part_of_speech: WordNet POS tag(s) to target. Accepts "n", "v", "a", "r",
      any iterable of those tags, or "any" to include all four.
    - rng: Optional RNG instance used for deterministic sampling.
    - seed: Optional seed if `rng` not provided.

    Determinism
    - Candidates collected in left-to-right order; no set() reordering.
    - Replacement positions chosen via rng.sample.
    - Synonyms sorted before rng.choice to fix ordering.
    - For each POS, the first synset containing alternate lemmas is used for stability.
    """
    effective_rate = resolve_rate(
        rate=rate,
        legacy_value=replacement_rate,
        default=0.1,
        legacy_name="replacement_rate",
    )

    ensure_wordnet()
    wordnet = _wordnet()

    active_rng: random.Random
    if rng is not None:
        active_rng = rng
    else:
        active_rng = random.Random(seed)

    target_pos = _normalize_parts_of_speech(part_of_speech)

    # Split but keep whitespace separators so we can rebuild easily
    tokens = re.split(r"(\s+)", text)

    # Collect indices of candidate tokens (even positions 0,2,.. are words given our split design)
    candidate_indices: list[int] = []
    candidate_metadata: dict[int, CandidateInfo] = {}
    for idx, tok in enumerate(tokens):
        if idx % 2 == 0 and tok and not tok.isspace():
            prefix, core_word, suffix = _split_token(tok)
            if not core_word:
                continue

            available_pos: NormalizedPartsOfSpeech = tuple(
                pos for pos in target_pos if wordnet.synsets(core_word, pos=pos)
            )
            if available_pos:
                candidate_indices.append(idx)
                candidate_metadata[idx] = CandidateInfo(
                    prefix=prefix,
                    core_word=core_word,
                    suffix=suffix,
                    parts_of_speech=available_pos,
                )

    if not candidate_indices:
        return text

    clamped_rate = max(0.0, effective_rate)
    max_replacements = int(len(candidate_indices) * clamped_rate)
    if max_replacements <= 0:
        return text

    # Choose which positions to replace deterministically via rng.sample
    replace_positions = active_rng.sample(candidate_indices, k=max_replacements)
    # Process in ascending order to avoid affecting later indices
    replace_positions.sort()

    for pos in replace_positions:
        metadata = candidate_metadata[pos]
        synonyms = _collect_synonyms(metadata.core_word, metadata.parts_of_speech)
        if not synonyms:
            continue

        replacement = active_rng.choice(synonyms)
        tokens[pos] = f"{metadata.prefix}{replacement}{metadata.suffix}"

    return "".join(tokens)


class Jargoyle(Glitchling):
    """Glitchling that swaps words with random WordNet synonyms."""

    def __init__(
        self,
        *,
        rate: float | None = None,
        replacement_rate: float | None = None,
        part_of_speech: PartOfSpeechInput = "n",
        seed: int | None = None,
    ) -> None:
        self._param_aliases = {"replacement_rate": "rate"}
        effective_rate = resolve_rate(
            rate=rate,
            legacy_value=replacement_rate,
            default=0.1,
            legacy_name="replacement_rate",
        )
        super().__init__(
            name="Jargoyle",
            corruption_function=substitute_random_synonyms,
            scope=AttackWave.WORD,
            seed=seed,
            rate=effective_rate,
            part_of_speech=part_of_speech,
        )


jargoyle = Jargoyle()


__all__ = ["Jargoyle", "dependencies_available", "ensure_wordnet", "jargoyle"]



import pytest

from glitchlings.util import string_diffs


def test_string_diffs_groups_consecutive_edits_and_skips_equals():
    result = string_diffs("kitten", "sitting")

    assert result == [
        [("replace", "k", "s")],
        [("replace", "e", "i")],
        [("insert", "", "g")],
    ]

    for group in result:
        assert group
        assert all(tag != "equal" for tag, *_ in group)


@pytest.mark.parametrize(
    "a,b,expected",
    [
        ("flaw", "lawn", [[("delete", "f", "")], [("insert", "", "n")]]),
        (
            "distance",
            "instance",
            [
                [("delete", "d", "")],
                [("insert", "", "n")],
            ],
        ),
    ],
)
def test_string_diffs_handles_multiple_edit_groups(a: str, b: str, expected: list[list[tuple[str, str, str]]]):
    assert string_diffs(a, b) == expected



import difflib

import pytest

from glitchlings import SAMPLE_TEXT, Typogre, summon
from glitchlings.main import (
    BUILTIN_GLITCHLINGS,
    DEFAULT_GLITCHLING_NAMES,
    MAX_NAME_WIDTH,
    build_parser,
    read_text,
    run_cli,
)


def invoke_cli(arguments: list[str]):
    parser = build_parser()
    args = parser.parse_args(arguments)
    exit_code = run_cli(args, parser)
    return exit_code


def render_expected_list_output() -> str:
    lines = []
    for key in DEFAULT_GLITCHLING_NAMES:
        glitchling = BUILTIN_GLITCHLINGS[key]
        scope = glitchling.level.name.title()
        order = glitchling.order.name.lower()
        lines.append(f"{glitchling.name:>{MAX_NAME_WIDTH}} — scope: {scope}, order: {order}")
    return "\n".join(lines) + "\n"


def render_expected_corruption(text: str, seed: int = 151) -> str:
    gaggle = summon(DEFAULT_GLITCHLING_NAMES, seed=seed)
    return gaggle(text)


def test_run_cli_lists_glitchlings(capsys):
    exit_code = invoke_cli(["--list"])
    captured = capsys.readouterr()
    assert exit_code == 0
    assert captured.out == render_expected_list_output()
    assert captured.err == ""


def test_run_cli_outputs_corrupted_sample_text(monkeypatch, capsys):
    class DummyStdin:
        def isatty(self):
            return True

        def read(self):
            raise AssertionError("stdin should not be read when running with --sample")

    monkeypatch.setattr("sys.stdin", DummyStdin())

    parser = build_parser()
    args = parser.parse_args(["--sample"])
    exit_code = run_cli(args, parser)
    captured = capsys.readouterr()
    assert exit_code == 0
    expected = render_expected_corruption(SAMPLE_TEXT, seed=args.seed)
    assert captured.out == expected + "\n"
    assert captured.err == ""


def test_run_cli_diff_mode(capsys):
    parser = build_parser()
    args = parser.parse_args(["--diff", "Hello, world!"])
    exit_code = run_cli(args, parser)
    captured = capsys.readouterr()
    assert exit_code == 0
    original = "Hello, world!"
    corrupted = render_expected_corruption(original, seed=args.seed)
    diff_lines = list(
        difflib.unified_diff(
            original.splitlines(keepends=True),
            corrupted.splitlines(keepends=True),
            fromfile="original",
            tofile="corrupted",
            lineterm="",
        )
    )
    if diff_lines:
        expected = "".join(f"{line}\n" for line in diff_lines)
    else:
        expected = "No changes detected.\n"
    assert captured.out == expected
    assert captured.err == ""


def test_run_cli_reads_text_from_file(tmp_path, capsys):
    input_text = "Corrupt me, glitchlings!"
    file_path = tmp_path / "input.txt"
    file_path.write_text(input_text, encoding="utf-8")
    parser = build_parser()
    args = parser.parse_args(["--file", str(file_path)])
    exit_code = run_cli(args, parser)
    captured = capsys.readouterr()
    assert exit_code == 0
    expected = render_expected_corruption(input_text, seed=args.seed)
    assert captured.out == expected + "\n"
    assert captured.err == ""


def test_read_text_reports_missing_file(tmp_path, capsys):
    parser = build_parser()
    missing = tmp_path / "missing.txt"
    args = parser.parse_args(["--file", str(missing)])
    with pytest.raises(SystemExit):
        read_text(args, parser)
    captured = capsys.readouterr()
    assert "No such file or directory" in captured.err
    assert str(missing) in captured.err


def test_read_text_requires_input(monkeypatch, capsys):
    parser = build_parser()
    args = parser.parse_args([])

    class DummyStdin:
        def isatty(self):
            return True

        def read(self):
            raise AssertionError("read should not be called when stdin is a tty")

    monkeypatch.setattr("sys.stdin", DummyStdin())

    with pytest.raises(SystemExit):
        read_text(args, parser)
    captured = capsys.readouterr()
    assert "No input text provided" in captured.err


def test_read_text_consumes_stdin(monkeypatch):
    parser = build_parser()
    args = parser.parse_args([])

    sentinel = "stdin payload"

    class DummyStdin:
        def isatty(self):
            return False

        def read(self):
            return sentinel

    monkeypatch.setattr("sys.stdin", DummyStdin())

    assert read_text(args, parser) == sentinel


def test_run_cli_configured_glitchling_matches_library(capsys):
    parser = build_parser()
    args = parser.parse_args(
        ["-g", "Typogre(rate=0.2)", "Hello there"]
    )

    exit_code = run_cli(args, parser)
    captured = capsys.readouterr()

    configured = Typogre(rate=0.2)
    expected = summon([configured], seed=args.seed)("Hello there")

    assert exit_code == 0
    assert captured.out == expected + "\n"
    assert captured.err == ""


def test_run_cli_rejects_positional_glitchling_arguments(capsys):
    parser = build_parser()
    args = parser.parse_args(["-g", "Typogre(0.2)", "payload"])

    with pytest.raises(SystemExit):
        run_cli(args, parser)

    captured = capsys.readouterr()
    assert "keyword arguments" in captured.err



"""Command line interface for summoning and running glitchlings."""

from __future__ import annotations

import argparse
import difflib
from pathlib import Path
import sys

from . import SAMPLE_TEXT
from .zoo import (
    Glitchling,
    Gaggle,
    BUILTIN_GLITCHLINGS,
    DEFAULT_GLITCHLING_NAMES,
    parse_glitchling_spec,
    summon,
)

MAX_NAME_WIDTH = max(len(glitchling.name) for glitchling in BUILTIN_GLITCHLINGS.values())


def build_parser() -> argparse.ArgumentParser:
    """Create and configure the CLI argument parser.

    Returns:
        argparse.ArgumentParser: The configured argument parser instance.
    """

    parser = argparse.ArgumentParser(
        description=(
            "Summon glitchlings to corrupt text. Provide input text as an argument, "
            "via --file, or pipe it on stdin."
        )
    )
    parser.add_argument(
        "text",
        nargs="?",
        help="Text to corrupt. If omitted, stdin is used or --sample provides fallback text.",
    )
    parser.add_argument(
        "-g",
        "--glitchling",
        dest="glitchlings",
        action="append",
        metavar="SPEC",
        help=(
            "Glitchling to apply, optionally with parameters like "
            "Typogre(rate=0.05). Repeat for multiples; defaults to all built-ins."
        ),
    )
    parser.add_argument(
        "-s",
        "--seed",
        type=int,
        default=151,
        help="Seed controlling deterministic corruption order (default: 151).",
    )
    parser.add_argument(
        "-f",
        "--file",
        type=Path,
        help="Read input text from a file instead of the command line argument.",
    )
    parser.add_argument(
        "--sample",
        action="store_true",
        help="Use the included SAMPLE_TEXT when no other input is provided.",
    )
    parser.add_argument(
        "--diff",
        action="store_true",
        help="Show a unified diff between the original and corrupted text.",
    )
    parser.add_argument(
        "--list",
        action="store_true",
        help="List available glitchlings and exit.",
    )
    return parser


def list_glitchlings() -> None:
    """Print information about the available built-in glitchlings."""

    for key in DEFAULT_GLITCHLING_NAMES:
        glitchling = BUILTIN_GLITCHLINGS[key]
        display_name = glitchling.name
        scope = glitchling.level.name.title()
        order = glitchling.order.name.lower()
        print(f"{display_name:>{MAX_NAME_WIDTH}} — scope: {scope}, order: {order}")


def read_text(args: argparse.Namespace, parser: argparse.ArgumentParser) -> str:
    """Resolve the input text based on CLI arguments.

    Args:
        args: Parsed arguments from the CLI.
        parser: The argument parser used for emitting user-facing errors.

    Returns:
        str: The text to corrupt.

    Raises:
        SystemExit: Raised indirectly via ``parser.error`` on failure.
    """

    if args.file is not None:
        try:
            return args.file.read_text(encoding="utf-8")
        except OSError as exc:
            filename = getattr(exc, "filename", None) or args.file
            reason = exc.strerror or str(exc)
            parser.error(f"Failed to read file {filename}: {reason}")

    if args.text:
        return args.text

    if not sys.stdin.isatty():
        return sys.stdin.read()

    if args.sample:
        return SAMPLE_TEXT

    parser.error(
        "No input text provided. Supply text as an argument, use --file, pipe input, or pass --sample."
    )
    raise AssertionError("parser.error should exit")


def summon_glitchlings(
    names: list[str] | None, parser: argparse.ArgumentParser, seed: int
) -> Gaggle:
    """Instantiate the requested glitchlings and bundle them in a ``Gaggle``."""

    if names:
        normalized: list[str | Glitchling] = []
        for specification in names:
            try:
                normalized.append(parse_glitchling_spec(specification))
            except ValueError as exc:
                parser.error(str(exc))
                raise AssertionError("parser.error should exit")
    else:
        normalized = DEFAULT_GLITCHLING_NAMES

    try:
        return summon(normalized, seed=seed)
    except ValueError as exc:
        parser.error(str(exc))
        raise AssertionError("parser.error should exit")



def show_diff(original: str, corrupted: str) -> None:
    """Display a unified diff between the original and corrupted text."""

    diff_lines = list(
        difflib.unified_diff(
            original.splitlines(keepends=True),
            corrupted.splitlines(keepends=True),
            fromfile="original",
            tofile="corrupted",
            lineterm="",
        )
    )
    if diff_lines:
        for line in diff_lines:
            print(line)
    else:
        print("No changes detected.")


def run_cli(args: argparse.Namespace, parser: argparse.ArgumentParser) -> int:
    """Execute the CLI workflow using the provided arguments.

    Args:
        args: Parsed CLI arguments.
        parser: Argument parser used for error reporting.

    Returns:
        int: Exit code for the process (``0`` on success).
    """

    if args.list:
        list_glitchlings()
        return 0

    text = read_text(args, parser)
    gaggle = summon_glitchlings(args.glitchlings, parser, args.seed)

    corrupted = gaggle(text)

    if args.diff:
        show_diff(text, corrupted)
    else:
        print(corrupted)

    return 0


def main(argv: list[str] | None = None) -> int:
    """Entry point for the ``glitchlings`` command line interface.

    Args:
        argv: Optional list of command line arguments. Defaults to ``sys.argv``.

    Returns:
        int: Exit code suitable for use with ``sys.exit``.
    """

    parser = build_parser()
    args = parser.parse_args(argv)
    return run_cli(args, parser)


if __name__ == "__main__":
    sys.exit(main())



import math
from typing import cast
from glitchlings import typogre, mim1c, reduple, rushmore, redactyl, scannequin


def _count_blocks(s: str, block_char: str = "\u2588") -> int:
    return s.count(block_char)


def test_mim1c_rate_bounds(sample_text):
    m = mim1c.clone()
    m.set_param("seed", 7)
    m.set_param("rate", 0.02)
    out = cast(str, m(sample_text))
    # Should change no more than ~2% of alnum characters
    alnum = [c for c in sample_text if c.isalnum()]
    paired_changes = sum(1 for a, b in zip(sample_text, out) if a != b and a.isalnum())
    if len(out) > len(sample_text):
        tail_alnum = sum(1 for c in out[len(sample_text):] if c.isalnum())
    else:
        tail_alnum = sum(1 for c in sample_text[len(out):] if c.isalnum())
    changed = paired_changes + tail_alnum
    assert changed <= int(len(alnum) * 0.02) + 2  # slack for discrete rounding


def test_mim1c_respects_banned_characters():
    m = mim1c.clone()
    m.set_param("seed", 2)
    m.set_param("rate", 1.0)
    m.set_param("banned_characters", ["ａ"])

    banned = {"a"}
    out = cast(str, m("a"))
    assert not any(char in banned for char in out)


def test_reduple_rate_increases_tokens():
    text = "a b c d e f g h"
    reduple.set_param("seed", 5)
    reduple.set_param("rate", 0.5)
    out = cast(str, reduple(text))
    assert len(out.split()) >= len(text.split())


def test_rushmore_rate_decreases_tokens():
    text = "a b c d e f g h"
    rushmore.set_param("seed", 5)
    rushmore.set_param("rate", 0.5)
    out = cast(str, rushmore(text))
    assert len(out.split()) <= len(text.split())


def test_rushmore_max_deletion_cap():
    text = "alpha beta gamma delta epsilon zeta eta theta"
    words = text.split()
    candidate_count = max(len(words) - 1, 0)

    for rate, seed in [(0.1, 3), (0.5, 11), (1.0, 17)]:
        rushmore.set_param("seed", seed)
        rushmore.set_param("rate", rate)
        out = cast(str, rushmore(text))

        removed = len(words) - len(out.split())
        allowed = min(candidate_count, math.floor(candidate_count * rate))
        assert removed <= allowed


def test_rushmore_preserves_leading_token_and_spacing():
    text = "Alpha, beta; gamma: delta epsilon zeta"
    seeds = (0, 3, 11, 21)
    rushmore.set_param("rate", 1.0)
    words = text.split()
    for seed in seeds:
        rushmore.set_param("seed", seed)
        out = cast(str, rushmore(text))
        leading = out.split()[0]
        original_core = "".join(ch for ch in words[0] if ch.isalnum())
        result_core = "".join(ch for ch in leading if ch.isalnum())
        assert leading
        assert result_core == original_core
        assert "  " not in out
        for marker in (" ,", " ;", " :", " ."):
            assert marker not in out
        assert out == out.strip()


def test_redactyl_replacement_char_and_merge():
    text = "alpha beta gamma"
    redactyl.set_param("seed", 2)
    redactyl.set_param("rate", 1.0)
    redactyl.set_param("replacement_char", "#")
    redactyl.set_param("merge_adjacent", True)
    out = cast(str, redactyl(text))
    assert set(out) <= {"#", " "}
    assert "# #" not in out  # merged


def test_scannequin_error_rate_increases_changes(sample_text):
    # count character diffs vs original
    def diff_count(a: str, b: str) -> int:
        return sum(1 for x, y in zip(a, b) if x != y) + abs(len(a) - len(b))

    scannequin.set_param("seed", 7)
    scannequin.set_param("rate", 0.005)
    low = cast(str, scannequin(sample_text))

    scannequin.set_param("seed", 7)
    scannequin.set_param("rate", 0.05)
    high = cast(str, scannequin(sample_text))

    assert diff_count(sample_text, high) >= diff_count(sample_text, low)



from __future__ import annotations


def resolve_rate(
    *,
    rate: float | None,
    legacy_value: float | None,
    default: float,
    legacy_name: str,
) -> float:
    """Return the effective rate while enforcing mutual exclusivity."""

    if rate is not None and legacy_value is not None:
        raise ValueError(
            f"Specify either 'rate' or '{legacy_name}', not both."
        )
    if rate is not None:
        return rate
    if legacy_value is not None:
        return legacy_value
    return default



import re
import random
from typing import Any

from .core import Glitchling, AttackWave
from ._rate import resolve_rate

try:
    from glitchlings._zoo_rust import reduplicate_words as _reduplicate_words_rust
except ImportError:  # pragma: no cover - compiled extension not present
    _reduplicate_words_rust = None


def _python_reduplicate_words(
    text: str,
    *,
    rate: float,
    rng: random.Random,
) -> str:
    """Randomly reduplicate words in the text.

    Parameters
    - text: Input text.
    - rate: Max proportion of words to reduplicate (default 0.05).
    - seed: Optional seed if `rng` not provided.
    - rng: Optional RNG; overrides seed.

    Notes
    - Preserves spacing and punctuation by tokenizing with separators.
    - Deterministic when run with a fixed seed or via Gaggle.
    """
    # Preserve exact spacing and punctuation by using regex
    tokens = re.split(r"(\s+)", text)  # Split but keep separators

    for i in range(0, len(tokens), 2):  # Every other token is a word
        if i >= len(tokens):
            break

        word = tokens[i]
        if not word or word.isspace():  # Skip empty or whitespace
            continue

        # Only consider actual words for reduplication
        if rng.random() < rate:
            # Check if word has trailing punctuation
            match = re.match(r"^(\W*)(.*?)(\W*)$", word)
            if match:
                prefix, core, suffix = match.groups()
                # Reduplicate with a space: "word" -> "word word"
                tokens[i] = f"{prefix}{core} {core}{suffix}"
            else:
                tokens[i] = f"{word} {word}"
    return "".join(tokens)


def reduplicate_words(
    text: str,
    rate: float | None = None,
    seed: int | None = None,
    rng: random.Random | None = None,
    *,
    reduplication_rate: float | None = None,
) -> str:
    """Randomly reduplicate words in the text.

    Falls back to the Python implementation when the optional Rust
    extension is unavailable.
    """

    effective_rate = resolve_rate(
        rate=rate,
        legacy_value=reduplication_rate,
        default=0.05,
        legacy_name="reduplication_rate",
    )

    if rng is None:
        rng = random.Random(seed)

    clamped_rate = max(0.0, effective_rate)

    if _reduplicate_words_rust is not None:
        return _reduplicate_words_rust(text, clamped_rate, rng)

    return _python_reduplicate_words(
        text,
        rate=clamped_rate,
        rng=rng,
    )


class Reduple(Glitchling):
    """Glitchling that repeats words to simulate stuttering speech."""

    def __init__(
        self,
        *,
        rate: float | None = None,
        reduplication_rate: float | None = None,
        seed: int | None = None,
    ) -> None:
        self._param_aliases = {"reduplication_rate": "rate"}
        effective_rate = resolve_rate(
            rate=rate,
            legacy_value=reduplication_rate,
            default=0.05,
            legacy_name="reduplication_rate",
        )
        super().__init__(
            name="Reduple",
            corruption_function=reduplicate_words,
            scope=AttackWave.WORD,
            seed=seed,
            rate=effective_rate,
        )

    def pipeline_operation(self) -> dict[str, Any] | None:
        rate = self.kwargs.get("rate")
        if rate is None:
            return None
        return {"type": "reduplicate", "reduplication_rate": float(rate)}



reduple = Reduple()


__all__ = ["Reduple", "reduple"]



"""Property-based tests covering core orchestration primitives."""

from __future__ import annotations

import math
import string

import importlib

import pytest

pytest.importorskip("hypothesis")

from hypothesis import HealthCheck, assume, given, settings, strategies as st

from glitchlings.zoo.core import AttackOrder, AttackWave, Gaggle, Glitchling

rushmore_module = importlib.import_module("glitchlings.zoo.rushmore")
typogre_module = importlib.import_module("glitchlings.zoo.typogre")


def _build_corruption(name: str, amplitude: int):
    """Create a deterministic corruption function driven by the provided RNG.

    The function appends a marker tied to the glitchling name along with a
    pseudo-random suffix that depends on the glitchling's RNG. This allows the
    tests to assert that derived seeds and ordering are both respected.
    """

    choices = (name + "xyz").replace("|", "_")

    def _corrupt(text: str, *, rng) -> str:
        if amplitude == 0:
            return f"{text}|{name}"
        suffix = "".join(rng.choice(choices) for _ in range(amplitude))
        return f"{text}|{name}:{suffix}"

    return _corrupt


@st.composite
def glitchling_specs(draw):
    name = draw(
        st.text(alphabet=string.ascii_letters + string.digits, min_size=1, max_size=8)
    )
    wave = draw(st.sampled_from(list(AttackWave)))
    order = draw(st.sampled_from(list(AttackOrder)))
    amplitude = draw(st.integers(min_value=0, max_value=4))
    return {"name": name, "wave": wave, "order": order, "amplitude": amplitude}


word_sequences = st.lists(
    st.text(alphabet=string.ascii_letters, min_size=1, max_size=12),
    min_size=2,
    max_size=12,
).map(lambda parts: " ".join(parts))


@given(
    master_seed=st.integers(min_value=-(2**63), max_value=2**63 - 1),
    specs=st.lists(glitchling_specs(), min_size=1, max_size=5, unique_by=lambda s: s["name"]),
)
def test_gaggle_ordering_and_determinism(master_seed, specs):
    """Gaggles should honour ordering guarantees and deterministic RNG use."""

    glitchlings = [
        Glitchling(
            name=spec["name"],
            corruption_function=_build_corruption(spec["name"], spec["amplitude"]),
            scope=spec["wave"],
            order=spec["order"],
        )
        for spec in specs
    ]

    gaggle = Gaggle(glitchlings, seed=master_seed)

    expected = [
        spec["name"]
        for spec in sorted(
            specs,
            key=lambda spec: (spec["wave"], spec["order"], spec["name"]),
        )
    ]
    actual = [g.name for g in gaggle.apply_order]
    assert actual == expected

    text = "payload"
    first_run = gaggle(text)
    second_run = Gaggle(glitchlings, seed=master_seed)(text)
    assert first_run == second_run


@given(
    left=st.tuples(
        st.integers(min_value=-(2**63), max_value=2**63 - 1),
        st.text(alphabet=string.ascii_letters + string.digits, min_size=0, max_size=12),
        st.integers(min_value=0, max_value=1024),
    ),
    right=st.tuples(
        st.integers(min_value=-(2**63), max_value=2**63 - 1),
        st.text(alphabet=string.ascii_letters + string.digits, min_size=0, max_size=12),
        st.integers(min_value=0, max_value=1024),
    ),
)
def test_derived_seeds_change_with_inputs(left, right):
    """Changing any component of the derivation tuple should alter the seed."""

    assume(left != right)
    assert Gaggle.derive_seed(*left) != Gaggle.derive_seed(*right)


@given(
    text=st.text(min_size=0, max_size=40),
    rate=st.floats(min_value=0.0, max_value=0.3),
    seed=st.integers(min_value=0, max_value=2**32 - 1),
)
def test_typogre_length_change_stays_within_bound(text: str, rate: float, seed: int) -> None:
    result = typogre_module.fatfinger(text, rate=rate, seed=seed)

    if not text:
        assert result == ""
        return

    max_changes = math.ceil(len(text) * rate)
    min_len = max(len(text) - max_changes, 0)
    max_len = len(text) + max_changes
    assert min_len <= len(result) <= max_len


@settings(suppress_health_check=[HealthCheck.filter_too_much, HealthCheck.too_slow])
@given(
    text=word_sequences,
    rate=st.floats(min_value=0.0, max_value=1.0),
    seed=st.integers(min_value=0, max_value=2**32 - 1),
)
def test_rushmore_preserves_first_token_and_respects_cap(text: str, rate: float, seed: int) -> None:
    words = text.split()

    result = rushmore_module.delete_random_words(text, rate=rate, seed=seed)
    result_words = result.split()
    assert result_words
    assert result_words[0] == words[0]

    candidate_count = len(words) - 1
    allowed = min(candidate_count, math.floor(candidate_count * rate))
    removed = len(words) - len(result_words)
    assert removed <= allowed




"""Integration helpers for the optional verifiers prime DLC."""

from __future__ import annotations

from collections.abc import Iterable, Sequence
from enum import Enum
from typing import Any, Callable

import verifiers as vf

from jellyfish import damerau_levenshtein_distance

try:
    from .huggingface import Dataset
except ModuleNotFoundError:  # pragma: no cover - optional dependency
    Dataset = object  # type: ignore[assignment]
else:
    if Dataset is None:  # pragma: no cover - optional dependency
        Dataset = object  # type: ignore[assignment]

from ..zoo import Gaggle, Glitchling, Mim1c, Typogre, summon


def _resolve_environment(env: str | vf.Environment) -> vf.Environment:
    """Return a fully-instantiated verifier environment."""

    if isinstance(env, str):
        env = vf.load_environment(env)

    if not isinstance(env, vf.Environment):
        raise TypeError("Invalid environment type")

    return env


def _resolve_columns(dataset: Dataset, columns: Sequence[str] | None) -> list[str]:
    """Identify which dataset columns should be corrupted."""

    available = set(dataset.column_names)

    if columns is not None:
        missing = sorted(set(columns) - available)
        if missing:
            missing_str = ", ".join(missing)
            raise ValueError(f"Columns not found in dataset: {missing_str}")
        return list(columns)

    for candidate in ("prompt", "question"):
        if candidate in available:
            return [candidate]

    sample = dataset[0] if len(dataset) else {}
    inferred = [
        name
        for name in dataset.column_names
        if isinstance(sample.get(name), str)
    ]

    if inferred:
        return inferred

    raise ValueError("Unable to determine which dataset columns to corrupt.")


class Difficulty(Enum):
    """Difficulty levels for tutorial environments."""

    Easy = 0.25
    Normal = 1.0
    Hard = 1.75
    Extreme = 3
    Impossible = 9


def tutorial_level(
    env: vf.Environment | str,
    seed: int = 151,
    difficulty: Difficulty = Difficulty.Normal,
) -> vf.Environment:
    """Create a low-corruption environment using tuned defaults."""

    tuned_mim1c = Mim1c(rate=0.01 * difficulty.value)
    tuned_typogre = Typogre(rate=0.025 * difficulty.value)

    return load_environment(
        env,
        glitchlings=[tuned_mim1c, tuned_typogre],
        seed=seed,
    )


def load_environment(
    env: str | vf.Environment,
    glitchlings: Iterable[str | Glitchling] | Glitchling | str | Gaggle | None = None,
    *,
    seed: int = 151,
    columns: Sequence[str] | None = None,
) -> vf.Environment:
    """Load an environment and optionally corrupt it with glitchlings."""

    environment = _resolve_environment(env)

    if glitchlings is None:
        return environment

    if isinstance(glitchlings, Gaggle):
        gaggle = glitchlings
    else:
        if isinstance(glitchlings, (Glitchling, str)):
            resolved = [glitchlings]
        else:
            resolved = list(glitchlings)

        gaggle = summon(resolved, seed=seed)

    dataset = environment.dataset
    corrupt_columns = _resolve_columns(dataset, columns)
    environment.dataset = gaggle.corrupt_dataset(dataset, corrupt_columns)
    return environment


def _as_gaggle(
    glitchlings: Iterable[str | Glitchling] | Glitchling | str | Gaggle,
    *,
    seed: int,
) -> Gaggle:
    """Coerce any supported glitchling specification into a :class:`Gaggle`."""

    if isinstance(glitchlings, Gaggle):
        return glitchlings

    if isinstance(glitchlings, (Glitchling, str)):
        resolved: Iterable[str | Glitchling] = [glitchlings]
    else:
        resolved = glitchlings

    return summon(list(resolved), seed=seed)


def _extract_completion_text(completion: Any) -> str:
    """Normalise a completion payload into a plain string."""

    if isinstance(completion, str):
        return completion

    if isinstance(completion, list) and completion:
        first = completion[0]
        if isinstance(first, dict) and "content" in first:
            return str(first["content"])
        return str(first)

    return str(completion)


def symmetric_damerau_levenshtein_similarity(
    _: Any,
    completion: Any,
    answer: str,
) -> float:
    """Return ``1 - (distance / max_len)`` using Damerau-Levenshtein distance."""

    completion_text = _extract_completion_text(completion)
    target = answer or ""
    denominator = max(len(completion_text), len(target), 1)
    distance = damerau_levenshtein_distance(completion_text, target)
    score = 1.0 - (distance / denominator)
    return max(0.0, min(1.0, score))


DEFAULT_CLEANUP_INSTRUCTIONS = (
    "You are a meticulous copy editor. Restore the provided text to its original form."
)


def echo_chamber(
    dataset_id: str,
    column: str,
    glitchlings: Iterable[str | Glitchling] | Glitchling | str | Gaggle,
    *,
    seed: int = 151,
    instructions: str = DEFAULT_CLEANUP_INSTRUCTIONS,
    reward_function: Callable[..., float] | None = None,
    split: str | None = None,
    **load_dataset_kwargs: Any,
) -> vf.Environment:
    """Create an Echo Chamber Prime environment from a Hugging Face dataset column.

    Args:
        dataset_id: Identifier of the Hugging Face dataset to load.
        column: Name of the column whose text should be glitched.
        glitchlings: Glitchling specifiers that will corrupt the prompts.
        seed: RNG seed forwarded to :func:`summon`.
        instructions: System instructions supplied to the environment prompts.
        reward_function: Optional callable used to score completions. Defaults to
            :func:`symmetric_damerau_levenshtein_similarity` when omitted.
        split: Optional dataset split to load.
        **load_dataset_kwargs: Extra keyword arguments forwarded to
            :func:`datasets.load_dataset`.
    """

    try:
        from datasets import Dataset as HFDataset, DatasetDict, load_dataset
    except ModuleNotFoundError as exc:  # pragma: no cover - optional dependency
        message = "datasets is required to build an echo chamber"
        raise ModuleNotFoundError(message) from exc

    hf_dataset: HFDataset | DatasetDict
    if split is None:
        hf_dataset = load_dataset(dataset_id, **load_dataset_kwargs)
        if isinstance(hf_dataset, DatasetDict):
            try:
                hf_dataset = next(iter(hf_dataset.values()))
            except StopIteration as exc:  # pragma: no cover - defensive
                raise ValueError("The specified dataset does not contain any splits") from exc
    else:
        hf_dataset = load_dataset(dataset_id, split=split, **load_dataset_kwargs)

    if isinstance(hf_dataset, DatasetDict):
        raise ValueError(
            "Specify which split to use when the dataset loads as a DatasetDict."
        )

    filtered_dataset = hf_dataset.filter(
        lambda row: row.get(column) is not None,
        load_from_cache_file=False,
    )

    source_column_names = list(filtered_dataset.column_names)

    def _build_prompt(row: dict[str, Any]) -> dict[str, Any]:
        text = str(row[column])
        prompt = [
            {"role": "system", "content": instructions},
            {"role": "user", "content": f"Corrupted text:\n{text}"},
        ]
        return {"prompt": prompt, "answer": text}

    base_dataset = filtered_dataset.map(
        _build_prompt,
        remove_columns=source_column_names,
        load_from_cache_file=False,
    )

    try:
        dataset_length = len(base_dataset)  # type: ignore[arg-type]
    except TypeError:
        preview_rows: list[dict[str, Any]]
        take_fn = getattr(base_dataset, "take", None)
        if callable(take_fn):
            preview_rows = list(take_fn(1))
        else:
            iterator = iter(base_dataset)
            try:
                first_row = next(iterator)
            except StopIteration:
                preview_rows = []
            else:
                preview_rows = [first_row]
        if not preview_rows:
            raise ValueError(
                f"Column '{column}' did not yield any textual entries in dataset '{dataset_id}'."
            )
    else:
        if dataset_length == 0:
            raise ValueError(
                f"Column '{column}' did not yield any textual entries in dataset '{dataset_id}'."
            )

    gaggle = _as_gaggle(glitchlings, seed=seed)
    glitched_dataset = gaggle.corrupt_dataset(base_dataset, ["prompt"])

    rubric_func = reward_function or symmetric_damerau_levenshtein_similarity
    rubric = vf.Rubric(funcs=[rubric_func], weights=[1.0])
    return vf.SingleTurnEnv(dataset=glitched_dataset, rubric=rubric)
